(***********************************************************************)
(*                                                                     *)
(*                          Sklml                                      *)
(*                                                                     *)
(*             Pierre Weis (Pierre.Weis@inria.fr)                      *)
(*             Francois Clement (Francois.Clement@inria.fr)            *)
(*             Quentin Carbonneaux (Quentin.Carbonneaux@inria.fr)      *)
(*                                                                     *)
(* This program is free software.                                      *)
(* (see the file LICENSE in the root directory of the distribution).   *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sklstream_old.ml,v 1.42 2010-11-22 10:29:52 qcarbonn Exp $ *)

(* Streams:
 No thunk on the wires (only plain values are transmitted)
 => no lazy values in the streams.
 Lazy evaluation in networks (meaning next elements is on demand)
 => lazy evaluation in the streams. *)

open Lazy;;

type 'a strict_lstream =
   | Cons of 'a * 'a lstream
   | Nil
and 'a lstream = 'a strict_lstream lazy_t;;

type 'a stream = 'a lstream lstream;;

type 'a t = 'a stream;;

let lnil = lazy Nil;;

let lazy_apply f x = lazy (
  f (force x)
);;
(** This function applies a function in a lazy evaluation scheme *)

let ( |$ ) f x = lazy_apply f x;;

let lazy_map lf x = lazy (
  (force lf) (force x)
);;

let ( |* ) lf x = lazy_map lf x;;

let wrap s = lazy (Cons (s, lnil));;
(* [wrap s] takes a simple lstream [s] and wrap it in a legal stream.
    It's kind of singleton, an order lower. *)

let singleton x = wrap (wrap x);;

let head s =
  let rec in_head ss =
    match force ss with
    | Cons (x, _) -> x
    | Nil -> raise Not_found in
  match force s with
  | Cons (ss, _) -> in_head ss
  | Nil -> raise Not_found
;;

(* let unsingleton s = *)
(*   let rec in_unsingleton ss acc = *)
(*     match force ss with *)
(*     | Cons (x, t) -> begin *)
(*       match acc with *)
(*       | Some _ -> in_unsingleton t acc *)
(*       | None -> in_unsingleton t (Some x) *)
(*     end *)
(*     | Nil -> *)
(*       match acc with *)
(*       | Some x -> x *)
(*       | None -> raise Not_found in *)
(*   let rec unsingleton s acc = *)
(*     match force s with *)
(*     | Cons (ss, t) -> begin *)
(*       match acc with *)
(*       | Some x -> *)
(*         unsingleton t (Some (in_unsingleton ss (Some x))) *)
(*       | None -> *)
(*         unsingleton t (Some (in_unsingleton ss None)) *)
(*     end *)
(*     | Nil -> *)
(*       match acc with *)
(*       | Some x -> x *)
(*       | None -> raise Not_found in *)
(*   unsingleton s None *)
(* ;; *)
(* (\** [unsingleton s] unfreezes the hole stream [s] and return the first *)
(*   element if found, otherwise raises [Not_found] *\) *)

let heads s =
  let rec in_heads s = lazy (
    match force s with
    | Cons (s, t) ->
      let rec get_head s acc =
        match force s with
        | Cons (x, t) -> begin
          match acc with
          | Some _ -> get_head t acc
          | None -> get_head t (Some x)
        end
        | Nil ->
          match acc with
          | Some x -> x
          | None -> raise Not_found in
      Cons (get_head s None, in_heads t)
    | Nil -> Nil) in
  wrap (in_heads s)
;;

(* [iter f s] iterates the function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [iter f s] performs the sequence
   [f x1; f x2; ...; f xn].
   Note: any call to [iter] completely unfreeezes its the stream argument. *)
let rec iter f s =
  let rec in_iter ss =
    match force ss with
    | Cons (x, t) -> f x; in_iter t
    | Nil -> () in
  match force s with
  | Cons (s, t) -> in_iter s; iter f t
  | Nil -> ()
;;

(* [foldl f e s] folds the binary function [f] on the stream [s]:
   if [s] is [x1; x2; ...; xn], then [foldl f e s] computes
   [f (... (f (f e x1) x2) ...) xn].
   Note: any call to [foldl] completely unfreeezes its stream argument. *)
let rec foldl f zero s =
  let rec in_fold t accu =
    match force t with
    | Cons (x, t) -> in_fold t (f (accu, x))
    | Nil -> wrap accu in

  (function
    | Cons (s, t) -> Cons (in_fold s zero, foldl f zero t)
    | Nil -> Nil) |$ s
;;

(* [map f s] maps the function [f] onto the stream [s]:
   if [s] is [x1; x2; ...; xn], then [map f s] returns the stream
   [f x1; f x2; ...; f xn]. *)
let rec map f s =
  let rec in_map s =
    (function
      | Cons (x, t) -> Cons (f x, in_map t)
      | Nil -> Nil) |$ s
  in
  (function
    | Cons (s, t) -> Cons (in_map s, map f t)
    | Nil -> Nil) |$ s
;;

(* Generate a lazy stream from a function. *)
let of_fun f =
  let rec of_fun f = lazy (
    match f () with
    | Some x -> Cons (x, of_fun f)
    | None -> Nil) in
  wrap (of_fun f)
;;

(* Generate a lazy stream from a list. *)
let of_list l =
  let rec of_list l = lazy (
    match l with
    | [] -> Nil
    | x :: t -> Cons (x, of_list t)) in
  wrap (of_list l)
;;

(* Generate the list of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_list s =
  let rec in_to_list accu t =
    match force t with
    | Cons (x, t) -> in_to_list (x :: accu) t
    | Nil -> accu in
  let rec to_list s acc =
    match force s with
    | Cons (ss, t) ->
      to_list t (in_to_list acc ss)
    | Nil -> List.rev acc in
  to_list s []
;;

(* Generate a stream from a vector. *)
let of_vect v =
  let in_of_vect = lazy (
    match v with
    | [| |] -> Nil
    | v ->
      let len = Array.length v in
      let rec loop i =
        if i >= len then Nil else
        Cons (v.(i), lazy (loop (i + 1))) in
      loop 0) in
  wrap (in_of_vect)
;;

(* Generate the vector of elements of a lazy stream.
   Warning: this function is eager, meaning that the stream argument is
   completely evaluated (if this is possible, otherwise the function loops for
   ever). *)
let to_vect s = Array.of_list (to_list s);;

(* Generate a lazy stream from a stream of lists. *)
let rec of_slist sl =
  let rec in_of_slist sl =
    (function
      | Cons (l, t) ->
        let rec loop l = lazy (
          match l with
          | [] -> Nil
          | h :: t -> Cons (h, loop t)) in
        Cons (loop l, in_of_slist t)
      | Nil -> Nil) |$ sl
  in

  (function
    | Cons (ss, _) -> force (in_of_slist ss)
    | Nil -> assert false) |$ sl
;;

let to_slist s =
  let rec in_to_slist ss =
    match force ss with
    | Cons (x, t) -> x :: in_to_slist t
    | Nil -> [] in

  let rec to_slist s =
    (function
      | Cons (ss, t) -> Cons (in_to_slist ss, to_slist t)
      | Nil -> Nil) |$ s
  in

  wrap (to_slist s)
;;

(* Generate a lazy stream from a stream of vectors. *)
let rec of_sarray vl =
  let rec in_of_sarray vl =
    (function
      | Cons (v, t) ->
        let to_stream v = lazy (
          match v with
          | [| |] -> Nil
          | v ->
            let len = Array.length v in
            let rec loop i =
              if i >= len then Nil else
              Cons (v.(i), lazy (loop (i + 1))) in
            loop 0) in
        Cons (to_stream v, in_of_sarray t)
      | Nil -> Nil) |$ vl
  in

  (function
    | Cons (ss, _) -> force (in_of_sarray ss)
    | Nil -> assert false) |$ vl
;;

let to_sarray s =
  let rec in_to_sarray ss =
    match force ss with
    | Cons (x, t) -> x :: in_to_sarray t
    | Nil -> [] in

  let rec to_sarray s =
    (function
      | Cons (ss, t) ->
        Cons (Array.of_list (in_to_sarray ss), to_sarray t)
      | Nil -> Nil) |$ s
  in

  wrap (to_sarray s)
;;

(* Internal functions *)

let lazy_tail s =
  (function
    | Nil -> assert false
    | Cons (_, t) -> Lazy.force t) |$ s
;;

let lazy_head s =
  (function
    | Nil -> assert false
    | Cons (h, _) -> Lazy.force h) |$ s
;;

(* Merges two lazy stream into a single stream with intermixed elements from
   both streams in a fair way (alternatively from one stream at a time).
   Note that the two streams must not have the same number of elements: in
   case one stream is exhausted, the result is the other (the longer one). *)
let rec join (s1, s2) =
  let rec in_join (s1, s2) =
    (function
      | Nil -> force s2
      | Cons (x, t) -> Cons (x, in_join (s2, t))) |$ s1
  in

  (function
    | Cons (ss1, t1) ->
      Cons
        (in_join (ss1, lazy_head s2),
         join (t1, lazy_tail s2))
    | Nil ->
      force s2) |$ s1
;;

let ffst x = force (fst x);;
let fsnd x = force (snd x);;

(* [fork s] separates the stream [s] into two streams in a fair way, even
   order elements are in one stream, odd order ones are in the other.
   (This means in particular, that if [s] is finite and has length [l], then both
   streams have length [l/2] when [l] is even, and respective length [l/2] and
   [l/2 + 1] when [l] is odd.) *)
let fork s =
  let rec in_fork ss =
    (function
      | Nil -> Nil, lnil
      | Cons (x, t) ->
        let pair = in_fork t in
        (Cons (x, fsnd |$ pair)),
        (fst |$ pair)) |$ ss
  in

  let rec fork s =
    (function
      | Nil -> Nil, Nil
      | Cons (ss, t) ->
        let pair = in_fork ss in
        Cons (fst |$ pair, fst |$ (fork t)),
        Cons (fsnd |$ pair, snd |$ (fork t))) |$ s
  in

  let pair = fork s in
  (fst |$ pair),
  (snd |$ pair)
;;

(* Not so experimental functions *)

let rec filter p s1 s2 =
  let rec in_filter ss1 ss2 =
    (fun x y ->
      match x, y with
      | Cons (b, t1), Cons (x, t2) ->
        if p b
        then Cons (x, in_filter t1 t2)
        else force (in_filter t1 t2)
      | Nil, Nil -> Nil
      | Cons _, Nil | Nil, Cons _ ->
        assert false) |$ ss1 |* ss2
  in

  (fun s1 s2 ->
    match s1, s2 with
    | Cons (ss1, t1), Cons (ss2, t2) ->
      Cons (in_filter ss1 ss2,
            filter p t1 t2)
    | Nil, Nil -> Nil
    | Nil, Cons _ | Cons _, Nil ->
      assert false) |$ s1 |* s2
;;

let with_head_tail f d s =
  match s with
  | Cons (h, t) -> f h t
  | Nil -> d
;;

let rec loop_core lb ls1 s2 =
  let rec in_loop_core b s1 s2 =
    (fun fb fs1 ->
      match fb, fs1 with
      | Cons (b, bs), Cons (x, xs) ->
        if b
        then Cons (x, in_loop_core bs xs s2)
        else
          force (
            (function
              | Cons (y, ys) ->
                Cons (y, in_loop_core bs xs ys)
              | Nil -> Nil) |$ s2
          )
      | (Nil | Cons _), (Nil | Cons _) ->
        assert false) |$ b |* s1
  in

  (function
    | Cons (ss2, t2) ->
      Cons
        ( (with_head_tail (* inject head *)
            (fun h t ->
              let b, s1 = force lb, force ls1 in
              Cons (h, in_loop_core (lazy_head b) (lazy_head s1) t ))
            Nil) |$ ss2
        , loop_core (lazy_tail |$ lb) (lazy_tail |$ ls1) t2
        )
    | Nil -> Nil) |$ s2
;;

let rec loop p f is =
  let rec los1 = lazy (f (Lazy.force lis1))
  and los2 = lazy (p (Lazy.force lis2))
  and lis1 = lazy (loop_core los2 los1 is)
  and lis2 = lazy (Lazy.force (los1)) in

  force (filter not |$ los2 |* los1)
;;

(* IO functions *)

type 'a packet =
   | Data of 'a
   | End
   | Start
;;
(** A ['a packet] is an object travelling on wires, it is used to
  serialize streams. *)

let send_receive_data (ic, oc) d =
  let kind_check d1 d2 =
    let check x y =
      if x = y then None else assert false in
    match d1 with
    | Start -> check Start d2
    | End -> check End d2
    | Data None -> check (Data None) d2
    | Data (Some _) ->
      let res =
        match d2 with
        | Start | End | Data None -> assert false
        | Data (Some y) -> Some y in
      res in

  IO.to_channel oc d true;
  let y = IO.of_channel ic in
  kind_check d y
;;

let of_to_channel io is =
  let rec in_of_to_channel ss =
    (function
      | Nil ->
        ignore (send_receive_data io (Data None));
        Nil
      | Cons (x, t) ->
        let y =
          match send_receive_data io (Data (Some x)) with
          | None -> assert false
          | Some y -> y in
        Cons (y, in_of_to_channel t)) |$ ss in

  let rec of_to_channel is =
    (function
      | Nil ->
        ignore (send_receive_data io (End));
        Nil
      | Cons (ss, t) ->
        Cons (in_of_to_channel ss, of_to_channel t)) |$ is
  in

  ignore (send_receive_data io (Start));
  of_to_channel is
;;

let of_channel ic = lazy (
  let rec in_of_channel ic = lazy (
    match IO.of_channel ic with
    | Data (Some x) -> Cons (x, in_of_channel ic)
    | Data None -> Nil
    | Start | End -> assert false) in
  let rec of_channel () =
    match IO.of_channel ic with
    | Data (Some x) ->
      Cons
        ( lazy (Cons (x, in_of_channel ic))
        , lazy (of_channel ())
        )
    | Data (None) ->
      Cons (lnil, lazy (of_channel ()))
    | Start -> assert false
    | End -> Nil in
  match IO.of_channel ic with
  | Start -> of_channel ()
  | Data _ | End -> assert false
);;

let to_channel s oc =
  let rec in_to_channel ss =
    match Lazy.force ss with
    | Cons (x, t) ->
      IO.to_channel oc (Data (Some x)) true;
      in_to_channel t
    | Nil -> IO.to_channel oc (Data None) true in
  let rec to_channel s =
    match Lazy.force s with
    | Cons (ss, t) ->
      in_to_channel ss;
      to_channel t
    | Nil -> IO.to_channel oc End true in
  IO.to_channel oc Start true;
  to_channel s
;;

let to_channel_core f s = lazy (
  let rec in_lazy_to_channel ss = lazy (
    match Lazy.force ss with
    | Cons (x, t) ->
      Cons (f (Data (Some x)), in_lazy_to_channel t)
    | Nil ->
      Cons (f (Data None), lazy Nil)) in
  let rec lazy_to_channel s = lazy (
    match Lazy.force s with
    | Cons (ss, t) ->
      Cons (
        in_lazy_to_channel ss,
        lazy_to_channel t)
    | Nil ->
      Cons (
        lazy (Cons (f End, lazy Nil)),
        lazy Nil)) in
  Cons (
    lazy (Cons (f Start, lazy Nil)),
    lazy_to_channel s)
);;

let stream_to_channel s oc =
  to_channel_core (fun x -> IO.to_channel oc x true) s
;;

(*let to_channel s oc = iter ignore (stream_to_channel s oc);;*)

let print_stream _data_printer _ppf _s =
assert false (* FIXME *)
;;

let rec eq_stream s1 s2 =
  let rec in_eq_stream ss1 ss2 =
    match force ss1, force ss2 with
    | Cons (x1, t1), Cons (x2, t2) ->
      x1 = x2 && in_eq_stream t1 t2
    | Nil, Nil -> true
    | (Cons _ | Nil), (Cons _ | Nil) ->
      false in
  match force s1, force s2 with
  | Cons (ss1, t1), Cons (ss2, t2) ->
    if in_eq_stream ss1 ss2 then eq_stream t1 t2 else false
  | Nil, Nil -> true
  | (Cons _ | Nil), (Cons _ | Nil) ->
    false
;;
